angular.module("beamng.apps")
/**
 * The base service, providing classes and functions that all Tachometer apps are likely to use
 */
.factory("TachoBase", [function() {
  // Mimic inner class behaviour by providing this service to those classes
  let outerThis = this;

  /**
   * A class representing a 2D vector
   */
  class Vector {
    constructor(x, y) {
      this.array = [x || 0, y || 0];
    }

    /**
     * Copies a vector and returns its copy
     * @param {Vector} v - the vector to copy
     * @returns {Vector} the copied vector instance
     */
    static copy(v) {
      return new Vector(v.array[0], v.array[1]);
    }

    /**
     * Reverse the signs of all elements in the vector
     */
    negate() {
      this.array.forEach(e => -e);
    }

    /**
     *
     * @param {Vector} vector
     * @param {Vector|number} amount
     * @param {function} func
     * @returns {Vector}
     */
    static mapOperations(vector, amount, func) {
      let result = new Vector(vector.array[0], vector.array[1]);

      if (amount instanceof Vector) {
        for (let i = 0; i < result.array.length; i++)
          result.array[i] = func(result.array[i], amount.array[i]);
      } else {
        result.array = result.array.map((e) => func(e, amount));
      }

      return result;
    }

    /**
     * Adds a vector from another or a scalar value
     * @param {Vector} vector - the vector to be added to
     * @param {Vector|number} amount - the vector/scalar to add
     * @returns {Vector} - the resulting vector
     */
    static add(vector, amount) {
      return Vector.mapOperations(vector, amount, (e, a) => e + a);
    }

    /**
     * Subtracts a vector from another or a scalar value
     * @param {Vector} vector - the vector to be subtracted from
     * @param {Vector|number} amount - the vector/scalar to subtract
     * @returns {Vector} - the resulting vector
     */
    static sub(vector, amount) {
      return Vector.mapOperations(vector, amount, (e, a) => e - a);
    }

    /**
     * Multiples a vector with another or a scalar value
     * @param {Vector} vector - the vector to be subtracted from
     * @param {Vector|number} amount - the vector/scalar to subtract
     * @returns {Vector} - the resulting vector
     */
    static mult(vector, amount) {
      return Vector.mapOperations(vector, amount, (e, a) => e * a);
    }

    /**
     * Divides a vector by another or a scalar value
     * @param {Vector} vector - the vector to be divided by
     * @param {Vector|number} amount - the vector/scalar to divide with
     * @returns {Vector} - the resulting vector
     */
    static div(vector, amount) {
      return Vector.mapOperations(vector, amount, (e, a) => e / a);
    }

    /**
     * Adds to this vector another vector from another or a scalar value
     * @param {Vector|number} amount - the vector/scalar to add
     */
    add(amount) {
      this.array = Vector.add(this, amount).array;
    }

    /**
     * Subtracts this vector by another or a scalar value
     * @param {Vector|number} amount - the vector/scalar to subtract
     */
    sub(amount) {
      this.array = Vector.sub(this, amount).array;
    }

    /**
     * Multiples this vector with another or a scalar value
     * @param {Vector|number} amount - the vector/scalar to subtract
     */
    mult(amount) {
      this.array = Vector.mult(this, amount).array;
    }

    /**
     * Divides this vector by another or a scalar value
     * @param {Vector|number} amount - the vector/scalar to divide with
     */
    div(amount) {
      this.array = Vector.div(this, amount).array;
    }

    /**
     * Normalises this vector
     */
    normalise() {
      let mag = this.mag();

      if (mag > 0)
        this.div(mag);
    }

    /**
     * Calculates and returns this vector's magnitude
     * @returns {number} the magnitude of this vector
     */
    mag() {
      return Math.sqrt(this.array.reduce((a, b) => Math.pow(a, 2) + Math.pow(b, 2)));
    }

    /**
     * Sets this vector's components to that of another's
     * @param {Vector} vector - the components to copy
     */
    set(vector) {
      for (let i = 0; i < this.array.length; i++)
        this.array[i] = vector.array[i];
    }

    setX(x) {
      this.array[0] = x;
    }

    setY(y) {
      this.array[1] = y;
    }

    addX(incX) {
      this.array[0] += incX;
    }

    addY(incY) {
      this.array[1] += incY;
    }

    x() {
      return this.array[0];
    }

    y() {
      return this.array[1];
    }
  }


  ///////////////////////
  // Drawing Functions //
  ///////////////////////

  /**
   * Draws a triangle whose base two vertices are positioned on a circle, and whose third vertex is
   * extended on another circle
   * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
   * @param {Vector} centre - the centre point of the UI
   * @param {number} centreAngle - the angle of the circle at the centre point of the triangle
   * @param {number} angleDev - the amount past the centre angle the triangle will extend to
   * @param {number} circleRadius - the circle's radius
   * @param {number} middleRadius - the radius the top-most vertex of the triangle will be placed on
   * @param {Style} styling - the styling to apply to the line
   */
  this.drawTriangleOnCircularBase = function(ctx, centre, centreAngle, angleDev, circleRadius, middleRadius, styling) {
    let p1 = this.getRelCircularPoint(centre, circleRadius, centreAngle - angleDev);
    let p2 = this.getRelCircularPoint(centre, circleRadius, centreAngle + angleDev);
    let p3 = this.getRelCircularPoint(centre, middleRadius, centreAngle);
    new Triangle(p1, p2, p3, styling).draw(ctx);
  }

  /////////////////////
  // Maths Functions //
  /////////////////////

  /**
   * Returns the coordinates of a point on a circle, whose centre is the origin (0,0)
   * @param {number} radius - the radius of the circle
   * @param {number} angle - the angle along the circle the point is
   * @returns {Vector} the point
   */
  this.getCircularPoint = function(radius, angle) {
    return new Vector(radius * Math.cos(angle), radius * Math.sin(angle));
  }

  /**
   * Returns the coordinates of a point on a circle, whose centre is specified
   * @param {Vector} relCentre - the centre of the circle to rotate around
   * @param {number} radius - the radius of the circle
   * @param {number} angle - the angle along the circle the point is
   * @returns {Vector} the point
   */
  this.getRelCircularPoint = function(relCentre, radius, angle) {
    return Vector.add(relCentre, this.getCircularPoint(radius, angle));
  }



  /**
   * Abstract
   * Represents an object is drawable to a canvas context object
   */
  class Drawable {
    constructor() {}

    /**
     * Draws the object to a canvas context object
     * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
     */
    draw(ctx) {
      console.error("draw() must be implemented by class!");
    }
  }

  /**
   * Abstract
   * Represents an object that can be moved
   */
  class Moveable extends Drawable {
    /**
     * Constructor
     * @param {Vector} [centre] - the centre point of the object. Defaults to (0,0)
     */
    constructor(centre) {
      super();
      this.centre = centre || new Vector();
    }

    /**
     * @param {Vector} centre - the new centre to set the object to
     */
    setCentre(centre) { this.centre = centre; };
  }

  /**
   * Abstract
   * Represents an object that can be rotated around a centre point
   */
  class Rotatable extends Moveable {
    /**
     * Constructor
     * @param {Vector} [centre] - the rotational centre. Defaults to (0,0)
     * @param {number} [angle] - the angle around the centre. Defaults to 0
     */
    constructor(centre, angle) {
      super(centre);
      this.angle = angle || 0;
    }

    /**
     * @param {number} angle - the angle to set
     */
    setAngle(angle) { this.angle = angle; };
  }

  /**
   * Represents a dual-colour scheme with a primary/default colour and a secondary/backup colour
   */
  class ColourPriority {
    /**
     * Constructor
     * @param {string} primary - the primary colour
     * @param {string} secondary - the secondary colour
     */
    constructor(primary, secondary) {
      this.primary = primary;
      this.secondary = secondary;
    }
  }

  /**
   * Represents a colour scheme for the UI
   */
  class ColourScheme {
    /**
     * Constructor
     * @param {string} base - the base colour, used for the main body
     * @param {string} background - the background elements' colour
     * @param {ColourPriority} text - the text colour
     * @param {ColourPriority} accent - the accent colour
     */
    constructor(base, background, text, accent) {
      this.BASE = base;
      this.BACKGROUND = background;
      // @type ColourPriority
      this.TEXT = text;
      // @type ColourPriority
      this.ACCENT = accent;
    }
  }

  /**
   * Represents a font for text displaying
   */
  class Font {
    /**
     * Constructor
     * @param {string} fontName - the font name
     * @param {number} fontSize - the font size
     * @param {boolean} isCentred - whether the text is centre
     */
    constructor(fontName, fontSize, isCentred) {
      this.fontName = fontName;
      this.fontSize = fontSize;
      this.isCentred = isCentred;
    }

    apply(ctx) {
      ctx.textAlign = this.isCentred ? "center" : "left";
      ctx.textBaseline = this.isCentred ? "middle" : "alphabetic";
      ctx.font = this.fontSize + "px " + this.fontName;
    }
  }

  /**
   * Represents text of a given font and position
   */
  class Text extends Drawable {
    /**
     * Constructor
     * @param {Font} font - the font to use for the text
     * @param {string} text - the text to display
     * @param {Vector} position - the position the text will be displayed at
     * @param {string} colour - the text colour
     */
    constructor(font, text, position, colour) {
      super();
      this.font = font;
      this.text = text;
      this.position = position;
      this.colour = colour;
    }

    /**
     * @Override
     * @inheritDoc
     */
    draw(ctx) {
      this.font.apply(ctx);
      ctx.fillStyle = this.colour;
      ctx.fillText(this.text, this.position.x(), this.position.y());
    }
  }

  /**
   * Represents a range of values between two values
   */
  class Range {
    /**
     * Constructor
     * The lower bound does not have to be smaller than the upper bound, and in which case, the
     * range is considered reversed
     * @param {number} lower - the lower bound of the range
     * @param {number} upper - the upper bound of the range
     */
    constructor(lower, upper) {
      this.lower = lower;
      this.upper = upper;
    }

    /**
     * Inverts the bounds such that the upper and lower bound swap
     * @returns {Range}
     */
    invert() {
      return new Range(this.upper, this.lower);
    }

    setLower(lower) { this.lower = lower; }
    setUpper(upper) { this.upper = upper; }

    /**
     * Pads the bounds outwards by a set amount. This lowers the lower bound by the set amount and
     * increases the upper bound by the set amount
     * @param {number} amount - the amount to pad
     */
    pad(amount) {
      this.lower -= amount;
      this.upper += amount;
    }

    /**
     * Clamps a value between this number range, such that a value lesser than the lower bound
     * will be set to the lower bound and vice versa for the upper bound. A value which lies
     * within the bounds is not affected
     * @param {number} value - the value to clamp
     * @returns {number} the clamped value
     */
    clampTo(value) {
      return Math.min(Math.max(value, this.lower), this.upper);
    }

    /**
     * Returns the value of the specified distance along this number range.
     * Example: {lower=5, upper=10}.getLinearRatio(0.5) -> 7.5
     * @param distance - the distance along the range
     * @returns {number} the value at {distance} along this range
     */
    getLinearRatio(distance) {
      return this.lower + distance * (this.upper - this.lower);
    }

    /**
     * @returns {number} the differences between the two bounds
     */
    getDifference() {
      return this.upper - this.lower;
    }

    clone() {
      return new Range(this.lower, this.upper);
    }
  }

  /**
   * Represents a style which can be applied to shapes and objects being drawn on a canvas
   */
  class Style {
    /**
     * Constructor
     * @param {string} strokeColour - the colour of the stroke. Can be either RGB/RGBA or colour
     * name
     * @param {number} lineWidth - the width of the line/outline
     * @param [filled] - whether the shape will be filled (fill colour will be the same as stroke
     * colour). Defaults to false
     */
    constructor(strokeColour, lineWidth, filled) {
      this.strokeColour = strokeColour;
      this.lineWidth = lineWidth;
      this.filled = filled || false;
    }

    /**
     * Returns a Style object of the specified colour that has a 1px stroke and is filled
     * @param {string} strokeColour - the colour of the stroke. Can be either RGB/RGBA or colour
     * name
     * @returns {Style} the style object created
     */
    static unitStrokeFill(strokeColour) {
      return new Style(strokeColour, 1, true);
    }

    /**
     * Returns a Style object that is filled with the specified colour and that does not have a
     * stroke applied to it
     * @param {string} fillColour - the fill colour. Can be either RGB/RGBA or colour name
     * @returns {Style} the style object created
     */
    static noStrokeFill(fillColour) {
      return new Style(fillColour, 0, true);
    }

    /**
     * Applies styling to a drawn-on canvas context and renders it
     * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
     */
    apply(ctx) {
      if (this.filled) {
        ctx.fillStyle = this.strokeColour;
        ctx.fill();
        return;
      }

      ctx.strokeStyle = this.strokeColour;
      ctx.lineWidth = this.lineWidth;
      ctx.stroke();
    }
  }


  /**
   * Abstract
   * Represents a drawable shape
   */
  class Shape extends Rotatable {
    /**
     * Constructor
     * @param {Style} styling - the styling to apply to the shape
     * @param {Vector} [rotationCentre] - the rotation point during rotation. Defaults to (0,0)
     * @param {Vector} [centre] - the centre point of the shape. Defaults to (0,0)
     */
    constructor(styling, centre, rotationCentre) {
      super(centre);
      this.rotationCentre = rotationCentre || new Vector();
      this.styling = styling;
    }

    /**
     * @Override
     * @inheritDoc
     */
    draw(ctx) {
      ctx.beginPath();
      this.applyRotation(ctx);
      this.drawContinue(ctx);
      this.applyStyling(ctx);
      Canvas.resetTransformationState(ctx);
    }

    /**
     * Applies the styling of this shape to a 2D canvas context
     * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to apply this
     * shapes styling to
     */
    applyStyling(ctx) {
      this.styling.apply(ctx);
    }

    /**
     * Sets the centre of rotation for this shape
     * @param centre
     */
    setRotationalCentre(centre) {
      this.rotationCentre = centre;
    }

    /**
     * Applies the rotation transformation of a canvas 2D context for drawing the rotated shape
     * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to rotate
     */
    applyRotation(ctx) {
      ctx.translate(this.rotationCentre.x(), this.rotationCentre.y());
      ctx.rotate(this.angle);
      ctx.translate(-this.rotationCentre.x(), -this.rotationCentre.y());
    }

    /**
     * Draws the arc on a canvas context object. Unlike draw(), this method does not begin nor end
     * the path for the given context object. Before calling this method, beginPath() should be
     * called to ensure correct rendering
     * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
     */
    drawContinue(ctx) {
      console.error("drawContinue() must be implemented by base class")
    }
  }

  /**
   * Represents a drawable arc
   */
  class Arc extends Shape {
    /**
     * Constructor
     * @param {number} radius - the radius of the arc
     * @param {Range} angles - the starting and end angles of the arc
     * @param {Style} styling - the styling to apply to the arc
     * @param {Vector} [centre] - the centre of the arc. Defaults to (0,0)
     */
    constructor(radius, angles, styling, centre) {
      super(styling, centre);
      this.radius = radius;
      this.angles = angles;
    }

    /**
     * @param {number} startAngle - the starting angle of the arc to set
     */
    setStartAngle(startAngle) { this.angles.setLower(startAngle); }

    /**
     * @param {number} endAngle - the ending angle of the arc to set
     */
    setEndAngle(endAngle) { this.angles.setUpper(endAngle); }

    getStartAngle() {
      return this.angles.lower;
    }

    getEndAngle() {
      return this.angles.upper;
    }

    /**
     * @Override
     * @inheritDoc
     */
    drawContinue(ctx) {
      ctx.arc(this.centre.x(), this.centre.y(), this.radius, this.angles.lower, this.angles.upper);
    }
  }

  /**
   * Represents a drawable circle
   */
  class Circle extends Arc {
    /**
     * Constructor
     * @param {number} radius - the radius of the circle
     * @param {Style} styling - the styling to apply to the circle
     * @param {Vector} [centre] - the centre of the circle. Defaults to (0,0)
     */
    constructor(radius, styling, centre) {
      super(radius, new Range(0, 2*Math.PI), styling, centre);
    }
  }

  class Wedge extends Arc {
    /**
     * Constructor
     * @param {number} radius - the radius of the wedge
     * @param {Range} angles - the starting and end angles of the wedge
     * @param {Style} styling - the styling to apply to the wedge
     * @param {Vector} [centre] - the centre of the wedge. Defaults to (0,0)
     */
    constructor(radius, angles, styling, centre) {
      super(radius, angles, styling, centre);
    }

    /**
     * @Override
     * @inheritDoc
     */
    draw(ctx) {
      ctx.beginPath();
      ctx.moveTo(this.centre.x(), this.centre.y());
      super.drawContinue(ctx);
      this.styling.apply(ctx);
    }
  }

  /**
   * Represents a drawable line
   */
  class Line extends Shape {
    /**
     * Constructor
     * @param {Vector} point1 - the start point
     * @param {Vector} point2 - the end point
     * @param {Style} styling - the styling to apply to the line
     */
    constructor(point1, point2, styling) {
      super(styling);
      this.point1 = point1;
      this.point2 = point2;
    }

    /**
     * @Override
     * @inheritDoc
     */
    setCentre(centre) {
      let oldCentre = this.centre;
      super.setCentre(centre);
      let translation = Vector.sub(this.centre, oldCentre);

      this.translate(translation);
    }

    translate(translation) {
      this.point1.add(translation);
      this.point2.add(translation);
    }

    /**
     * @Override
     * @inheritDoc
     */
    drawContinue(ctx) {
      ctx.moveTo(this.point1.x(), this.point1.y());
      ctx.lineTo(this.point2.x(), this.point2.y());
    }
  }

  /**
   * Represents a drawable triangle
   */
  class Triangle extends Line {
    /**
     * Constructor
     * @param {Vector} point1 - the first vertex
     * @param {Vector} point2 - the second vertex
     * @param {Vector} point3 - the third vertex
     * @param {Style} styling - the styling to apply to the triangle
     */
    constructor(point1, point2, point3, styling) {
      super(point1, point2, styling);
      this.point3 = point3;

      // Calculate and set centre
      this.setCentre(this.calculateCentrePoint());
    }

    /**
     * @Override
     * @inheritDoc
     */
    translate(translation) {
      super.translate(translation);
      this.point3.add(translation);
    }

    /**
     * Constructs and returns an isosceles triangle.
     * @param {number} baseWidth - the width of the base
     * @param {number} height - the height
     * @param {Style} styling - the styling to apply to the triangle
     * @param {Vector} centre - the point to put the triangle's centre to
     */
    static newIsosceles(baseWidth, height, styling, centre) {
      let p1 = new Vector( baseWidth/2, 0);
      let p2 = new Vector(-baseWidth/2, 0);
      let p3 = new Vector(0, height);

      let triangle = new Triangle(p3, p1, p2, styling);
      triangle.setCentre(centre);
      return triangle;
    }

    /**
     * Calculates the centre point of the triangle
     * @returns {Vector}
     */
    calculateCentrePoint() {
      let sum = new Vector();
      sum.add(this.point1);
      sum.add(this.point2);
      sum.add(this.point3);
      return Vector.div(sum, 3);
    }

    /**
     * @Override
     * @inheritDoc
     */
    drawContinue(ctx) {
      super.drawContinue(ctx);
      ctx.lineTo(this.point3.x(), this.point3.y());
    }
  }

  /**
   * Represent a rectangle
   */
  class Rectangle extends Shape {
    /**
     * Constructor
     * @param {number} width - the width of the rectangle
     * @param {number} height - the height of the rectangle
     * @param styling - the styling to apply
     */
    constructor(width, height, styling) {
      super(styling);
      this.width = width;
      this.height = height;
    }

    /**
     * @Override
     * @inheritDoc
     */
    drawContinue(ctx) {
      ctx.rect(this.centre.x(), this.centre.y(), this.width, this.height);
    }
  }

  /**
   * Represents a rectangle with a rounded bottom-right corner
   */
  class RoundCornerRectangleBR extends Shape {
    /**
     * Constructor
     * @param {number} arcRadius - the rounded part's arc radius
     * @param {Range} arcAngles - the rounded part's arc angle range
     * @param {Vector} arcCentre - the rounded part's arc centre point
     * @param {number} leftX - the leftmost point on the x dimension
     * @param {Style} styling - the style to apply to this shape
     */
    constructor(arcRadius, arcAngles, arcCentre, leftX, styling) {
      super(styling);
      this.arc = new Arc(arcRadius, arcAngles, styling, arcCentre);
      this.leftX = leftX;
    }

    drawContinue(ctx) {
      this.arc.drawContinue(ctx);
      ctx.lineTo(this.leftX, outerThis.getRelCircularPoint(this.arc.centre, this.arc.radius, this.arc.getEndAngle()).y());
      ctx.lineTo(this.leftX, outerThis.getRelCircularPoint(this.arc.centre, this.arc.radius, this.arc.getStartAngle()).y());
    }
  }


  /**
   * Represents a completion bar in ma arc shape
   */
  class ArcBar extends Drawable {
    /**
     * Constructor
     * By default, the bar is 100% complete
     * @param {number} radius - the radius of the arc
     * @param {Range} angles - the starting and end angles of the arc
     * @param {Style} styling - the styling to apply to the arc
     * @param {Vector} [centre] - the centre of the arc. Defaults to (0,0)
     */
    constructor(radius, angles, styling, centre) {
      super();
      this.bgArc = new Arc(radius, angles, new Style("rgba(255, 255, 255, 0.25)", styling.lineWidth, false), centre);
      this.mainArc = new Arc(radius, angles.clone(), styling, centre);
    }

    /**
     * @param {number} perc - the percentage full to set the bar to be (between [0,1])
     */
    setPercentageAlong(perc) {
      let angle = this.bgArc.angles.getLinearRatio(perc);
      this.mainArc.setEndAngle(angle);
    }

    /**
     * @Override
     * @inheritDoc
     */
    draw(ctx) {
      ctx.lineCap = "round";
      this.bgArc.draw(ctx);
      this.mainArc.draw(ctx);
      ctx.lineCap = "butt";
    }
  }


  /**
   * Represents a canvas object that can be drawn on
   */
  class Canvas {
    /**
     * Constructor
     * @param {HTMLDocument} document - the document object this canvas is on
     * @param {string} id - the canvas's identifier string
     */
    constructor(document, id) {
      this.element = document.getElementById(id);
      this.ID = id;

      this.setupCanvas();
    }

    /**
     * Ensures proper HD scaling of text to eliminate blurry text and geometry
     */
    setupCanvas() {
      // Ensure correct HD resolution on the canvas (fix for blurry text)
      this.element.width = this.element.getBoundingClientRect().width;
      this.element.height = this.element.getBoundingClientRect().height;
      this.show(false);
    }

    /**
     * Shows or hides the canvas
     * @param {boolean} show - whether to show or hide the canvas
     */
    show(show) {
      this.element.style.opacity = show ? "1" : "0";
    }

    /**
     * @returns {CanvasRenderingContext2D} the 2D context of the canvas
     */
    get2DContext() {
      return this.element.getContext("2d");
    }

    /**
     * Clears the canvas of all content drawn on it
     */
    clear() {
      let ctx = this.get2DContext();
      ctx.clearRect(0, 0, this.element.width, this.element.height);
    }

    /**
     * Sets a listener to a canvas for the left and right click action by a mouse
     * @param {function} leftFunc - the function to perform on a left click
     * @param {function} rightFunc - the function to perform on a right click
     */
    setupCanvasClick(leftFunc, rightFunc) {
      this.element.addEventListener("mousedown", function(event) {
        switch (event.button) {
          case 0:
            leftFunc(event);
            break;
          case 2:
            rightFunc(event);
            break;
        }
      });
    }

    /**
     * Resets the transformation of a canvas's 2D context
     * @param {CanvasRenderingContext2D} ctx - the context to reset the transformation of
     */
    static resetTransformationState(ctx) {
      ctx.setTransform(1, 0, 0, 1, 0, 0);
    }
  }


  /**
   * Abstract
   * Represents a gauge pointer
   */
  class GaugePointer extends Rotatable {
    constructor() {
      super();
    }
  }

  /**
   * Represents a gauge pointer as a collection of the centre dial, minor pointer and major pointer
   */
  class SimplePointer extends GaugePointer {
    /**
     * Constructor
     * @param {Shape} majorPointer - the major part of the pointer
     * @param {Shape} minorPointer - the minor part of the pointer (points to the inverse angle)
     * @param {Shape} pointerDial - the pointer dial at the centre
     */
    constructor(majorPointer, minorPointer, pointerDial) {
      super();
      this.majorPointer = majorPointer;
      this.minorPointer = minorPointer;
      this.pointerDial = pointerDial;
    }

    /**
     * @Override
     * @inheritDoc
     */
    setAngle(angle) {
      super.setAngle(angle);
      this.majorPointer.setAngle(angle);
      this.minorPointer.setAngle(this.getInvAngle());
    }

    /**
     * @Override
     * @inheritDoc
     */
    setCentre(centre) {
      super.setCentre(centre);
      this.majorPointer.setCentre(centre);
      this.minorPointer.setCentre(centre);
    }

    /**
     * Sets the centre of rotation for the pointers
     * @param centre
     */
    setRotationalCentre(centre) {
      this.majorPointer.setRotationalCentre(centre);
      this.minorPointer.setRotationalCentre(centre);
    }

    /**
     * @Override
     * Draws the pointer pieces to a canvas
     */
    draw(ctx) {
      this.drawPointers(ctx);
      this.pointerDial.draw(ctx);
    }

    /**
     * Draws the pointers
     * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
     */
    drawPointers(ctx) {
      this.majorPointer.draw(ctx);
      this.minorPointer.draw(ctx);
    }

    /**
     * returns the angle 180deg added to the current angle. Note that the value returned will
     * always be +180deg from the current, regardless of whether this makes the angle bigger than
     * 360deg
     * @returns {number} the angle 180deg plus the current angle, in radians
     */
    getInvAngle() { return this.angle + Math.PI; }
  }

  /**
   * Represents a pointer object, wherein an accent is displayed over the pointer
   */
  class AccentPointer extends GaugePointer {
    /**
     * Constructor
     * The angle this pointer points at is initially set to 0
     * @param {SimplePointer} bgPointer - the background pointer
     * @param {SimplePointer} accentPointer - the accent pointer
     * @param {Circle} pointerDial - the centre dial part of the pointers
     */
    constructor(bgPointer, accentPointer, pointerDial) {
      super();
      this.bgPointer = bgPointer;
      this.accentPointer = accentPointer;
      this.pointerDial = pointerDial;
    }

    /**
     * @Override
     * @inheritDoc
     */
    setAngle(angle) {
      super.setAngle(angle);

      // An angle of 0 means the pointer is at the bottom -> -Math.PI/2
      angle -= Math.PI/2;
      this.accentPointer.setAngle(angle);
      this.bgPointer.setAngle(angle);
    }

    /**
     * Sets the centre of rotation for the pointers
     * @param centre
     */
    setRotationalCentre(centre) {
      this.accentPointer.setRotationalCentre(centre);
      this.bgPointer.setRotationalCentre(centre);
    }

    /**
     * @Override
     * Draws the pointer pieces to a canvas.
     * Note that pointerDial overrides the pointer dials specified in the pointer
     * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
     */
    draw(ctx) {
      this.bgPointer.drawPointers(ctx);
      this.accentPointer.drawPointers(ctx);
      this.pointerDial.draw(ctx);
    }
  }


  /**
   * Abstract
   * Represents an element that sits on top of the background of a gauge
   */
  class GaugeFaceElement extends Moveable {
    /**
     * Constructor
     * @param {number} numElements - the number of elements
     * @param {number} valueSpacing - the value spacing between adjacent numbers
     * @param {Range} angleRange - the starting and ending angle range of the elements
     */
    constructor(numElements, valueSpacing, angleRange) {
      super();
      this.numElements = numElements;
      this.valueSpacing = valueSpacing;
      this.angleRange = angleRange;
      // -1 as element 0 starts at angleRange.lower and element N ends at angleRange.upper
      // so include upper bound as element N (not N+1)
      this.angleSpacing = angleRange.getDifference() / (numElements-1);
      this.elements = [];
    }

    /**
     * Abstract
     * Generates the elements
     * @return {?} the elements
     */
    generateElements() {
      console.error("generateElements() must be implemented in subclass!")
    }
  }

  /*
   * Represents the gauge's ticks that line the gauge
   */
  class GaugeTicks extends GaugeFaceElement {
    /**
     * Constructor
     * @param {number} numTicks - the number of ticks on gauge
     * @param {number} spacing - the spacing between two ticks, in terms of value
     * @param {number} width - the stroke width of each tick
     * @param {number} majorSpacing - the number of ticks between two major ticks
     * @param {ColourPriority} colour - the colour to apply to the ticks
     * @param {Range} tickRadii - the inner and outer radii of the ticks
     * @param {Range} angles - the angle of the first tick and the angle of the last tick
     * @param {boolean} [drawEndTicks] - whether to draw the ticks at either of the ends. Defaults
     * to false
     * @param {number} [minTickVal] - the lowest tick's value. Defaults to 0
     */
    constructor(numTicks, spacing, width, majorSpacing, colour, tickRadii, angles, drawEndTicks, minTickVal) {
      super(numTicks, spacing, angles);
      this.majorSpacing = majorSpacing;
      this.width = width;
      this.colour = colour;
      this.tickRadii = tickRadii;
      this.drawEndTicks = drawEndTicks || false;
      this.minTickVal = minTickVal || 0;

      this.elements = this.generateElements();
    }

    /**
     * @Override
     * @inheritDoc
     */
    setCentre(centre) {
      let oldCentre = this.centre;
      super.setCentre(centre);
      let translation = Vector.sub(this.centre, oldCentre);

      // Translate the ticks to the new centre point
      for (let tick of this.elements) {
        tick[0].add(translation);
        tick[1].add(translation);
      }
    }

    /**
     * @Override
     * @inheritDoc
     * Generates a list of ticks as pairs of radii
     * @returns {[[Vector, Vector]]} the ticks, as pairs of inner and outer radii
     */
    generateElements() {
      let ticks = [];
      let currAngle = this.angleRange.lower;

      for (let i = 0; i < this.numElements; i++) {
        let isMajor = i % this.majorSpacing !== 0;
        let isEndTick = i === 0 || i === this.numElements-1;

        // Only draw end ticks if wanted
        if (!isEndTick || this.drawEndTicks)
          ticks.push(this.addTick(currAngle, isMajor));

        currAngle += this.angleSpacing;
      }

      return ticks;
    }

    /**
     * Returns an RPM tick, as a pair of radii, of the given inputs. The radii returned are the
     * outer and inner radii of the tick.
     * @param {number} angle - the angle this tick sits at around the RPM
     * @param {boolean} isMinor - whether this tick is representing a minor point on the gauge (ie.
     * a thousand on an RPM gauge)
     * @returns {[Vector, Vector]} the radii of the tick
     */
    addTick(angle, isMinor) {
      let r1 = this.getTickRadius(false);
      let r2 = this.getTickRadius(true);
      // Half size for minor ticks
      if (isMinor)
        r2 += (r1-r2)/2;

      return [
        outerThis.getRelCircularPoint(this.centre, r1, angle),
        outerThis.getRelCircularPoint(this.centre, r2, angle)
      ];
    }

    /**
     * Returns the radius on the RPM gauge a tick will start/end at
     * @param {boolean} isInner - whether to return the inner or outer radius
     * @returns {number}
     */
    getTickRadius(isInner) {
      return isInner ? this.tickRadii.lower : this.tickRadii.upper;
    }

    /**
     * Returns the accent colour of a tick based on whether it has been covered by the current
     * value on the gauge
     * @param {boolean} isValPastTick - whether the tick being drawn is covered by the current
     * value on the gauge
     * @returns {string} the accent colour
     */
    getAccentColour(isValPastTick) {
      return isValPastTick ? this.colour.primary : this.colour.secondary;
    }

    /**
     * Returns the value at the specified tick index
     * @param {number} index - the tick index
     * @returns {number} the vale at that tick index
     */
    getValueAtIndex(index) {
      // If no ticks are being drawn, index 0 is actually index 1
      let actualIndex = this.drawEndTicks ? index : index + 1;
      return this.minTickVal + actualIndex * this.valueSpacing;
    }

    /**
     * Draws ticks on the gauge as a disconnected series of lines. If the current value is greater
     * than the tick, the secondary accent colour will be applied; else, the primary accent will
     * be applied
     * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
     * @param {number} currValue - the current value the angle on the gauge corresponds to
     */
    drawTicks(ctx, currValue) {
      for (let i = 0; i < this.elements.length; i++) {
        let tick = this.elements[i];
        // If current value is past this tick, make it a different colour
        let colour = this.getAccentColour(currValue >= this.getValueAtIndex(i));

        new Line(tick[0], tick[1], new Style(colour, this.width))
            .draw(ctx);
      }
    };
  }

  /**
   * Represents a gauges numbering system that lines the gauge and displays that gauge's value at
   * set intervals
   */
  class GaugeNumbers extends GaugeFaceElement {
    /**
     * Constructor
     * @param {number} numNumbers - the number of gauge numbers
     * @param {number} spacing - the numerical spacing between adjacent gauge numbers
     * @param {Range} angles - the start and end angles the gauge numbers will start and end at
     * @param {Font} font - the font to use for displaying the gauge numbers
     * @param {number} radius - the radius of the circle the gauge numbers will wrap around from
     * @param {string} textColour - the colour of the gauge numbers
     * @param {number} [startNum] - the number to start at. Defaults to 0
     */
    constructor(numNumbers, spacing, angles, font, radius, textColour, startNum) {
      super(numNumbers, spacing, angles);
      this.font = font;
      this.radius = radius;
      this.textColour = textColour;
      this.startNum = startNum || 0;

      this.elements = this.generateElements();
    }

    /**
     * @Override
     * @inheritDoc
     */
    setCentre(centre) {
      let oldCentre = this.centre;
      super.setCentre(centre);
      let translation = Vector.sub(this.centre, oldCentre);

      // Translate the ticks to the new centre point
      for (let number of this.elements)
        number.position.add(translation);
    }

    /**
     * @Override
     * @inheritDoc
     * @returns {[Text]} the numbers generated
     */
    generateElements() {
      let numbers = [];
      let angle = this.angleRange.lower;

      for (let i = 0; i < this.numElements; i++) {
        let value = this.startNum + (i * this.valueSpacing);
        numbers.push(this.generateNumber(Math.trunc(value), angle));

        angle += this.angleSpacing;
      }

      return numbers;
    }

    generateNumber(number, angle) {
      let position = outerThis.getRelCircularPoint(this.centre, this.radius, angle);
      return new Text(this.font, number.toString(), position, this.textColour);
    }

    /**
     * @Override
     * @inheritDoc
     */
    draw(ctx) {
      this.elements.forEach((e) => e.draw(ctx));
    }
  }

  /**
   * Represents the arc(s) that wrap around the outside of a gauge
   */
  class GaugeArcs extends Drawable {
    /**
     * Constructor
     * @param {Arc} mainArc - the main arc which will represent the current value
     * @param {Arc} [endArc] - the arc that sits at the end. This is optional, and an example would
     * be the redline buffer for an engine or boost gauge
     * @param {Arc} [startArc] - the arc that sits at the start. This is optional
     */
    constructor(mainArc, endArc, startArc) {
      super();
      this.mainArc = mainArc;
      this.endArc = endArc;
      this.startArc = startArc;
    }

    /**
     * Sets the centres for all arcs
     * @param {Vector} centre - the centre position to set
     */
    setCentre(centre) {
      if (this.endArc)
        this.endArc.setCentre(centre);
      if (this.startArc)
        this.startArc.setCentre(centre);

      this.mainArc .setCentre(centre);
    }

    /**
     * @Override
     * @inheritDoc
     * It should be noted that the start and end arcs (if they exist) have higher drawing priority
     * over the main arc and so will always be shown on top
     */
    draw(ctx) {
      this.mainArc.draw(ctx);

      if (this.endArc)
        this.endArc.draw(ctx);
      if (this.startArc)
        this.startArc.draw(ctx);
    }
  }

  /**
   * Represents a gauge with a pointer and ticks that number the main parts of the gauge. This
   * gauge also has an arc around the outside that fills an area between the lowest angle and
   * the angle the current value's angle.
   */
  class Gauge extends Drawable {
    /**
     * Constructor
     * @param {GaugePointer} pointer - the pointer on the gauge pointing to the current value
     * @param {GaugeTicks} gaugeTicks - the ticks on the gauge
     * @param {GaugeArcs} gaugeArcs - tha arc(s) on the gauge showing the current value and any
     * start and end 'padding arcs. @see the GaugeArcs constructor for more information
     * @param {GaugeNumbers} gaugeNumbers - the numbers labelling the gauge
     * @param {Vector} centre - the centre point of the gauge
     */
    constructor(pointer, gaugeTicks, gaugeArcs, gaugeNumbers, centre) {
      super();
      this.pointer = pointer;
      this.gaugeTicks = gaugeTicks;
      this.gaugeArcs = gaugeArcs;
      this.gaugeNumbers = gaugeNumbers;

      this.setupCentres(centre);
    }

    /**
     * Sets the centres for the gauge's arc and pointer
     */
    setupCentres(centre) {
      this.pointer     .setCentre(centre);
      this.gaugeArcs   .setCentre(centre);
      this.gaugeTicks  .setCentre(centre);
      this.gaugeNumbers.setCentre(centre);
    }

    /**
     * @param {number} currValueAngle - the new angle of the current value on the gauge
     */
    setAngle(currValueAngle) {
      // Ensure it does not go past either end of the gauge
      currValueAngle = this.gaugeTicks.angleRange.clampTo(currValueAngle);

      this.pointer.setAngle(currValueAngle);
      this.gaugeArcs.mainArc.setEndAngle(currValueAngle);
    }

    /**
     * Draws the object to screen
     * @param ctx
     * @param currValue
     */
    draw(ctx, currValue) {
      // Note that order is important for correct layering
      this.gaugeArcs   .draw(ctx);
      this.gaugeNumbers.draw(ctx);
      this.pointer     .draw(ctx);
      this.gaugeTicks.drawTicks(ctx, currValue);
    }
  }

  /**
   * Abstract
   * A builder for creating a Gauge object
   */
  class GaugeBuilder {
    /**
     * Constructor
     */
    constructor() {
      this.pointer = null;
      this.gaugeTicks = null;
      this.gaugeArcs = null;
      this.gaugeNumbers = null;
      this.centre = null;
    }

    setCentre(centre) {
      this.centre = centre;
      this.pointer.setRotationalCentre(centre);
      return this;
    }

    /**
     * Builds the Gauge instance and return it
     * @returns {Gauge} the gauge object created. {@code null} if any of the components required to
     * build the gauge are not created
     */
    build() {
      if (!this.pointer || !this.gaugeTicks || !this.gaugeArcs || !this.gaugeNumbers || !this.centre) {
        console.err("One or more components for this gauge have not been setup. Ensure that all " +
            "parts have been built before calling build()!");
        return null;
      }

      return new Gauge(
          this.pointer,
          this.gaugeTicks,
          this.gaugeArcs,
          this.gaugeNumbers,
          this.centre
      );
    }
  }


  /**
   * Abstract
   * Represents the base app that all modded apps will use/extend
   */
  class BaseApp {
    /**
     * Constructor
     * {@param bngApi} is required as it cannot be requested here by Angular's injector for some
     * reason :/
     * @param {number} id - the ID of this app
     * @param {number} baseRadius - the radius of the outermost part of the app
     * @param {String} fontName - the name of the font to use
     * @param {[String]} units - the list of units that can be used for this app
     * @param {HTMLDocument} document - the document element the app uses
     * @param {?} bngApi - the bngApi service
     * @param {?} UiUnits - the UiUnits service
     */
    constructor(id, baseRadius, fontName, units, document, bngApi, UiUnits) {
      this.ID = id;
      this.FONT_NAME = fontName;
      this.UNITS = units;
      this.radius = baseRadius;
      this.CENTRE = new Vector(baseRadius, baseRadius);

      // Default to first
      this.unit = units[0];

      // A list of states the UI can be in
      this.states = Object.freeze({"preInit":1, "init":2, "loaded":3, "hidden":4});
      // The current state
      this.state = this.states.init;

      this.gauge = null;

      this.COLOUR_SCHEMES = outerThis.COLOUR_SCHEMES;
      this.colourScheme = this.COLOUR_SCHEMES[0];

      this.CANVAS    = new Canvas(document, this.ID + "_canvas");
      this.BG_CANVAS = new Canvas(document, this.ID + "_bgCanvas");

      this.UiUnits = UiUnits;

      this.initBGCanvas(bngApi);
    }

    /**
     * Sets up the background canvas, adding an event handler for the mouse action
     */
    initBGCanvas(bngApi) {
      let thisApp = this;

      this.CANVAS.setupCanvasClick(
          function(_event) {
            // Change unit
            thisApp.cycleUnit();
            thisApp.recreate(bngApi, true);
          },
          function(_event) {
            // Change colour scheme
            thisApp.cycleColourScheme();
            thisApp.recreate(bngApi, true);
          }
      );
    }

    /**
     * Recreates the gauge and re-draws the background canvas
     */
    recreate(bngApi, resave) {
      // Re-create gauge to reflect unit changes
      this.createGauge();
      // Show unit on background canvas
      this.drawBackgroundCanvas();

      if (resave)
        // Save this colour scheme and unit as default
        this.save(bngApi);
    }

    /**
     * Changes the current unit to the next available option
     */
    cycleUnit() {
      this.unit = BaseApp.cycleArray(this.unit, this.UNITS);
    }

    /**
     * Changes the current colour scheme to the next available option
     */
    cycleColourScheme() {
      this.colourScheme = BaseApp.cycleArray(this.colourScheme, this.COLOUR_SCHEMES);
    }

    /**
     * Saves the current colour scheme and unit to a file in the user's directory. Note that this
     * only saves the indexes with respect to the COLOUR_SCHEMES array and not the actual instance
     */
    save(bngApi) {
      let data = {
        unit: this.UNITS.indexOf(this.unit),
        colourScheme: this.COLOUR_SCHEMES.indexOf(this.colourScheme)
      }

      bngApi.engineLua("extensions.ui_tachometers.save('" + this.ID + "', " + bngApi.serializeToLua(data) + ")");
    }

    /**
     * Loads the settings file from the user's directory. If the file does not exist, this method
     * does nothing; else, it loads the unit and colour scheme to this app
     * @param {?} bngApi - the BeamNG API for interfacing with Lua
     */
    load(bngApi) {
      bngApi.engineLua("extensions.ui_tachometers.load('" + this.ID + "')", (response) => {
        if (!response)
          return;

        this.unit = this.UNITS[response.unit];
        this.colourScheme = this.COLOUR_SCHEMES[response.colourScheme];

        this.recreate(bngApi, false);
      });
    }

    /**
     * Draws to the main canvas
     */
    draw() {
      this.CANVAS.show(true);
      this.CANVAS.clear();
    }

    /**
     * Draws the app to the background canvas
     */
    drawBackgroundCanvas() {
      this.BG_CANVAS.show(true);
      this.BG_CANVAS.clear();
    }

    /**
     * Hides the app by hiding both canvases
     */
    hide() {
      this.CANVAS.show(false);
      this.BG_CANVAS.show(false);
      this.state = this.states.hidden;
    }


    /**
     * Returns the next value in an array, given a current value, looping around to the start if
     * the value is at the end of the array
     * @template T
     * @param {T} currValue - the current value
     * @param {[T]} array - the array containing a set of values, including {@param currValue}
     * @returns {T} the next value in {@param array}, succeeding {@param currValue}
     */
    static cycleArray(currValue, array) {
      // Find next index, looping back around to 0 if necessary
      let index = array.indexOf(currValue);
      index = (index+1) % array.length;

      return array[index];
    }

    /**
     * Resets the filter to prevent lingering filters
     */
    static resetFilter(ctx) {
      ctx.filter = "none";
    }

    /**
     * Reloads the UI, hiding it after 50ms. If instantReload is true, this is done
     * instantly
     * @param {BoostApp} app - the app to reload
     * @param {boolean} instantReload - whether to instantly reload instead of waiting
     */
    static reload(app, instantReload) {
      if (!app)
        return;

      let time = instantReload ? 0 : 50;
      setTimeout(() => {
        // It is unknown whether this vehicle will need this app, so hide it provisionally and
        // it is does, the app will re-show itself when updating occurs
        app.hide();
        app.state = app.states.init;
      }, time);
    }


    /**
     * Creates the gauge element
     */
    createGauge() {
      console.error("createGauge() is abstract and must be implemented in the base class!");
    }
  }

  /**
   * Abstract
   * Represents a base Tachometer modded app
   */
  class TachometerApp extends BaseApp {
    /**
     * Constructor
     * @param {number} id - the ID of this app
     * @param {number} baseRadius - the radius of the outermost part of the app
     * @param {String} fontName - the name of the font to use
     * @param {Range} rpmAngleRange - the RPM angle range of the gauge
     * @param {HTMLDocument} document - the document element the app uses
     * @param {?} bngApi - the bngApi service
     * @param {?} UiUnits - the UiUnits service
     */
    constructor(id, rpmAngleRange, fontName, baseRadius, document, bngApi, UiUnits) {
      super(id, baseRadius, fontName, ["metric", "imperial"], document, bngApi, UiUnits);
      this.RPM_ANGLE_RANGE = rpmAngleRange;

      this.lightsState = 0;
      this.currGearIndex = 0;
      this.speed = 0;
      this.currRPM = 0;
      this.maxRPM = 0;
      this.redLineRPM = 0;
      this.pBrakeEngaged = false;
      this.engineTemp = 0;
      this.fuel = 0;
    }

    /**
     * Updates the variables in the app's fields. All parameters are optional, and all are updated
     * via the respective field's setter function
     * @param {number} [currGearIndex] - the current gear's index the vehicle is in
     * @param {number} [speed] - the current speed of the vehicle
     * @param {number} [currRPM] - the current RPM of the engine
     * @param {number} [lightsState] - the state of the lights (0=off, 1=low, 2=high)
     * @param {boolean} [pBrakeEngaged] - whether the parking brake is engaged or not
     * @param {number} [engineTemp] - the temperature of the engine's oil
     * @param {number} fuel - the percentage of fuel left
     */
    updateVariables(currGearIndex, speed, currRPM, lightsState, pBrakeEngaged, engineTemp, fuel) {
      this.setCurrGearIndex(currGearIndex);
      this.setSpeed(speed);
      this.setCurrRPM(currRPM);
      this.setLightsState(lightsState);
      this.setParkingBrake(pBrakeEngaged);
      this.setEngineTemp(engineTemp);
      this.setFuel(fuel);
    }

    /**
     * Initialises the app's fields
     * @param {number} currGearIndex - the current gear's index the vehicle is in
     * @param {number} speed - the current speed of the vehicle
     * @param {number} currRPM - the current RPM of the engine
     * @param {number} maxRPM - the maximum RPM of the engine
     * @param {number} lightsState - the state of the lights (0=off, 1=low, 2=high)
     * @param {number} fuel - the percentage of fuel left
     */
    initFields(currGearIndex, speed, currRPM, maxRPM, lightsState, fuel) {
      this.updateVariables(currGearIndex, speed, currRPM, lightsState, fuel);
      this.setMaxRPM(maxRPM);
      // Round to next thousand
      this.setRedLineRPM(Math.floor(maxRPM / 1000 + 1) * 1000);

      this.createGauge();
    }

    setCurrGearIndex(gearIndex) {
      if (typeof gearIndex !== "undefined")
        this.currGearIndex = gearIndex;
    }

    setSpeed(speed) {
      if (typeof speed !== "undefined")
        this.speed = speed;
    }

    setCurrRPM(RPM) {
      if (typeof RPM !== "undefined")
        this.currRPM = RPM;
    }

    setLightsState(lightsState) {
      if (typeof lightsState !== "undefined" && lightsState >= 0 && lightsState <= 2)
        this.lightsState = lightsState;
    }

    setMaxRPM(maxRPM) {
      if (typeof maxRPM !== "undefined")
        this.maxRPM = maxRPM;
    }

    setRedLineRPM(redLineRPM) {
      if (typeof redLineRPM !== "undefined")
        this.redLineRPM = redLineRPM;
    }

    setParkingBrake(pBrakeEngaged) {
      if (typeof pBrakeEngaged !== "undefined")
        this.pBrakeEngaged = pBrakeEngaged;
    }

    setEngineTemp(engineTemp) {
      if (typeof engineTemp !== "undefined")
        this.engineTemp = engineTemp;
    }

    setFuel(fuel) {
      if (typeof fuel !== "undefined")
        this.fuel = fuel;
    }

    /**
     * @returns {number} the angle at which redline RPM occurs on the RPM gauge
     */
    getRedlineRPMAngle() {
      return this.RPM_ANGLE_RANGE.getLinearRatio(this.maxRPM / this.redLineRPM);
    }

    /**
     * Converts a gear index into its string name.
     * Mapping:
     *   < -1  -> R2...Rn
     *   = -1  -> R
     *   =  0  -> N
     *   >  1  -> 1...n
     * @param {number} gearIndex - the gear's index
     * @returns {string} the string name for that gear index
     */
    static gearIndexToName(gearIndex) {
      let gearName;

      if (gearIndex === 0) {
        gearName = "N";
      } else if (gearIndex < 0) {
        // Add index of reverse gear if more than one exists
        if (gearIndex < -1)
          gearName = "R" + Math.abs(gearIndex);
        else
          gearName = "R";
      } else {
        gearName = gearIndex.toString();
      }

      return gearName;
    }
  }

  /**
   * Abstract
   * Represents a base Boost modded app
   */
  class BoostApp extends BaseApp {
    /**
     * Constructor
     * @param {number} id - the ID of this app
     * @param {number} baseRadius - the radius of the outermost part of the app
     * @param {String} fontName - the name of the font to use
     * @param {Range} boostAngleRange - the boost angle range of the gauge
     * @param {HTMLDocument} document - the document element the app uses
     * @param {?} bngApi - the bngApi service
     * @param {?} UiUnits - the UiUnits service
     */
    constructor(id, boostAngleRange, fontName, baseRadius, document, bngApi, UiUnits) {
      super(id, baseRadius, fontName, ["inHg", "bar", "psi", "kPa"], document, bngApi, UiUnits);
      this.BOOST_ANGLE_RANGE = boostAngleRange;

      this.maxBoost = 0;
      this.currBoost = 0;
    }

    /**
     * Initialises the app's fields.
     * Note that both values should be their 'raw' values, so that they can be converted to specific
     * units later
     * @param {number} currBoost - the current boost of the turbo
     * @param {number} maxBoost - the maximum possible boost value
     */
    initFields(currBoost, maxBoost) {
      this.setMaxBoost(maxBoost);
      this.setCurrBoost(currBoost);

      this.createGauge();
    }

    setMaxBoost(maxBoost) {
      if (typeof maxBoost !== "undefined")
        this.maxBoost = maxBoost;
    }

    setCurrBoost(currBoost) {
      if (typeof currBoost !== "undefined")
        this.currBoost = currBoost;
    }
  }


  this.Vector = Vector;
  this.ColourScheme = ColourScheme;
  this.Font = Font;
  this.Text = Text;
  this.Range = Range;
  this.Style = Style;

  this.Arc = Arc;
  this.Circle = Circle;
  this.Wedge = Wedge;
  this.Line = Line;
  this.Triangle = Triangle;
  this.RoundCornerRectangleBR = RoundCornerRectangleBR;

  this.ArcBar = ArcBar;

  this.SimplePointer = SimplePointer;
  this.AccentPointer = AccentPointer;
  this.GaugeTicks = GaugeTicks;
  this.GaugeNumbers = GaugeNumbers;
  this.GaugeArcs = GaugeArcs;
  this.Gauge = Gauge;
  this.GaugeBuilder = GaugeBuilder;

  this.BaseApp = BaseApp;
  this.TachometerApp = TachometerApp;
  this.BoostApp = BoostApp;



  ////////////////////////
  // The Colour Schemes //
  ////////////////////////

  /**
   * Represents an RGB colour
   */
  class ColourRGB {
    constructor(r, g, b) {
      if (typeof r === "object" && !g && !b)
        this.rgb = r.map((x) => x);
      else
        this.rgb = [r, g, b];
    }

    toString() {
      return this.rgb[0] + ", " + this.rgb[1] + ", " + this.rgb[2];
    }
  }

  function createColourScheme(rgb1, rgb2, rgb3) {
    // This is constant for the colour schemes
    const GREY_SCALE_SCHEME = {
      BASE: "rgb(40, 40, 40)",
      SECONDARY_TEXT: "rgb(255, 255, 255)",
      ACCENT: "rgb(0, 0, 0)"
    }

    rgb2 = rgb2 || rgb1;
    rgb3 = rgb3 || rgb1;

    return new ColourScheme(
      GREY_SCALE_SCHEME.BASE,
      "rgba(" + rgb1.toString() + ", 0.75)",
      new ColourPriority("rgb(" + rgb2.toString() + ")", GREY_SCALE_SCHEME.SECONDARY_TEXT),
      new ColourPriority(GREY_SCALE_SCHEME.ACCENT, "rgb(" + rgb3.toString() + ")"),
    )
  }

  function generateColourSchemes() {
    let colours = {
      red      : new ColourRGB(255,   0,   0),
      orange   : new ColourRGB(255, 127,   0),
      green    : new ColourRGB(  0, 255,   0),
      blue     : new ColourRGB(  0, 100, 255),
      yellow   : new ColourRGB(255, 255,   0),
      lightblue: new ColourRGB(  0, 255, 255),
      purple   : new ColourRGB(150,   0, 255),
      grey     : new ColourRGB(100, 100, 100),
      white    : new ColourRGB(255, 255, 255)
    }

    let colourSchemes = [];

    // Generate tri-coloured colour schemes
    // BeamNG
    colourSchemes.push(createColourScheme(colours.grey, colours.orange, colours.orange));
    // Fire
    colourSchemes.push(createColourScheme(colours.red, colours.yellow, colours.orange));
    // Arsenic
    colourSchemes.push(createColourScheme(colours.green, colours.yellow, colours.green));
    // Crystal
    colourSchemes.push(createColourScheme(colours.blue, colours.lightblue, colours.purple));

    // Generate uni-coloured colour schemes
    for (let key in colours) {
      // Ignore grey as too dark
      if (key === "grey")
        continue;

      colourSchemes.push(createColourScheme(colours[key]));
    }


    return colourSchemes;
  }

  this.COLOUR_SCHEMES = generateColourSchemes();


  return this;
}])