angular.module("beamng.apps")
/**
 * The base service for the future variant apps
 */
.factory("ModVelocityBase", ["TachoBase", function(TachoBase) {
  let outerThis = this;

  /**
   *
   */
  class GaugeBuilder extends TachoBase.GaugeBuilder {
    /**
     * Constructor
     * @param {number} baseRadius - the radius of the gauge at its maximum point
     * @param {ColourScheme} colourScheme - the colour scheme to apply to the gauge
     * @param {Range} angleRange - the range of angles of the dial's arc overall
     */
    constructor(baseRadius, colourScheme, angleRange) {
      super();
      // @type number
      this.baseRadius = baseRadius;
      // @type ColourScheme
      this.colourScheme = colourScheme;
      // @type Range
      this.angleRange = angleRange;
    }

    /**
     * @Override
     * @param {Number} dialRadius - the radius of the middle dial
     * @param {Vector} centre - the centre point of the dial
     * @param {Number} majorLength - the length of the pointer on the major side
     * @returns {GaugeBuilder} the GaugeBuilder instance
     */
    createPointer(dialRadius, centre, majorLength) {
      let accent = this.colourScheme.ACCENT;
      let primaryStyle   = TachoBase.Style.unitStrokeFill(accent.primary);
      let secondaryStyle = TachoBase.Style.unitStrokeFill(accent.secondary);

      this.pointer = new TachoBase.AccentPointer(
          new TachoBase.SimplePointer(
              TachoBase.Triangle.newIsosceles(majorLength/4, majorLength, primaryStyle, centre),
              TachoBase.Triangle.newIsosceles(majorLength/4, majorLength/2, primaryStyle, centre),
              null
          ),
          new TachoBase.SimplePointer(
              TachoBase.Triangle.newIsosceles(majorLength/12, majorLength, secondaryStyle, centre),
              TachoBase.Triangle.newIsosceles(majorLength/12, majorLength/2, secondaryStyle, centre),
              null
          ),
          new TachoBase.Circle(dialRadius, new TachoBase.Style(accent.primary, 5, true), centre)
      )
      return this;
    }

    /**
     * Creates the GaugeTicks instance for the gauge
     * @param {number} numTicks - the number of ticks to generate
     * @param {number} tickSpacing - the spacing between two ticks, in terms of value
     * @param {number} arcWidth - the width of the arc
     * @param {number} minTickVal - the lowest tick's value. Defaults to 0
     * @returns {GaugeBuilder} the GaugeBuilder instance
     */
    createTicks(numTicks, tickSpacing, arcWidth, minTickVal) {
      this.gaugeTicks = new TachoBase.GaugeTicks(
          numTicks,
          tickSpacing,
          2,
          4,
          this.colourScheme.ACCENT,
          // +-2 to not go over bounds
          new TachoBase.Range(this.baseRadius - arcWidth + 2, this.baseRadius - 2),
          this.angleRange,
          false,
          minTickVal
      );
      return this;
    }

    /**
     * Creates the GaugeNumbers instance for the gauge
     * @param {number} numNumbers - the number of numbers that will fill the major points on the
     * tick gauge
     * @param {number} fontSize - the size of the font for the numbers
     * @param {number} numbersRadius - the radius of the arc that the numbers will be spaced around
     * @param {number} spacing - the spacing between two numbers
     * @param {number} [startNum] - the starting number. Defaults to 0
     * @returns {GaugeBuilder} the GaugeBuilder instance
     */
    createNumbers(numNumbers, fontSize, numbersRadius, spacing, startNum) {
      this.gaugeNumbers = new TachoBase.GaugeNumbers(
          numNumbers,
          spacing,
          this.angleRange,
          new TachoBase.Font(outerThis.FONT_NAME, fontSize, true),
          numbersRadius,
          this.colourScheme.TEXT.secondary,
          startNum
      );
      return this;
    }
  }


  const VT_WIDTH = {
    // The current RPM arc that goes around the outside
    RPM_ARC: 30,
    // The interior's stroke between itself and the RPM background
    INNER_STROKE: 5,
    // The width of each tick on the RPM gauge
    RPM_TICK_WIDTH: 2,
    // The width of each bar (fuel and temp gauge)
    BARS: 6
  };
  // The background to the RPM thousand numbers
  VT_WIDTH.RPM_BG = 80 - VT_WIDTH.RPM_ARC;

  const VT_RADII = {
    // The radius of the whole UI
    BASE: 150,
    // The radius of the inner circle/background part
    BG_INNER: 90,
    // The radius of the pointer's dial
    RPM_DIAL_RADIUS: 30
  };
  // The radius of the stroke that separates the RPM numbers' background and the inner
  // background piece
  VT_RADII.RPM_ARC_BG = VT_RADII.BG_INNER + VT_WIDTH.RPM_BG / 2;
  // The radius of the RPM arc that goes around the outside of the gauge
  VT_RADII.RPM_ARC = VT_RADII.BASE - VT_WIDTH.RPM_ARC / 2;
  // The radius of the inner (stroked) part of the RPM arc
  VT_RADII.RPM_INNER = VT_RADII.BASE - VT_WIDTH.RPM_ARC;
  // The radius of the inner background piece
  VT_RADII.MIDDLE = VT_RADII.BG_INNER - VT_WIDTH.INNER_STROKE * 2;
  // The radius the RPM numbers are displayed around
  VT_RADII.RPM_NUMBERS = VT_RADII.RPM_INNER - (VT_RADII.RPM_INNER - VT_RADII.BG_INNER + VT_WIDTH.INNER_STROKE) / 2;
  // The radius of the temperature bar
  VT_RADII.TEMP = VT_RADII.MIDDLE;
  VT_RADII.FUEL = VT_RADII.TEMP - 10;


  /**
   * The concrete class representation of this app
   */
  class VelocityTachoApp extends TachoBase.TachometerApp {
    // The width of stroke effects for various components
    static WIDTH = VT_WIDTH;

    /*
     * The centre radius values for various components.
     * It should be noted that the radii values are with respect to an arc that they represent.
     * So, a radius value is not the radius of that element, but the radius of the arc around the
     * centre of the UI it will be drawn as.
     * Any stroke effect adds thickness to a radius
     */
    static RADII = VT_RADII;

    // The RPM spacing between two ticks
    static TICK_SPACING = 250;


    constructor(id, document, bngApi, UiUnits) {
      super(id, outerThis.GAUGE_RANGE, outerThis.FONT_NAME, VelocityTachoApp.RADII.BASE, document, bngApi, UiUnits);
      this.needsRedraw = false;

      this.oilTempBar = this.createOilTempBar();
      this.fuelBar = this.createFuelBar();
    }

    /**
     * @Override
     * @inheritDoc
     */
    updateVariables(currGearIndex, speed, currRPM, lightsState, pBrakeEngaged, engineTemp, fuel) {
      let prevState = this.collateBGVars();
      super.updateVariables(currGearIndex, speed, currRPM, lightsState, pBrakeEngaged, engineTemp, fuel);
      let newState = this.collateBGVars();

      // If any variables for the background canvas have changed, the background canvas
      // requires a re-draw
      if (!prevState.every(function(v, i) { return v === newState[i]}))
        this.needsRedraw = true;
    }

    /**
     * @returns {[*]} the list of all variables that drawn on the background canvas. This
     * includes the gear index, the lights' state and the parking brake
     */
    collateBGVars() {
      return [this.currGearIndex, this.lightsState, this.pBrakeEngaged];
    }

    /**
     * @Override
     * @inheritDoc
     */
    createGauge() {
      // Create concrete class of base class builder
      class TachoVelocityGaugeBuilder extends outerThis.GaugeBuilder {
        constructor(baseRadius, colourScheme, angleRange) {
          super(baseRadius, colourScheme, angleRange);
        }

        /**
         * Creates the GaugeArcs instance for the gauge
         * @param arcRadius
         * @param arcWidth
         * @param redlineAngle
         * @returns {TachoVelocityGaugeBuilder} the GaugeBuilder instance
         */
        createArcs(arcRadius, arcWidth, redlineAngle) {
          this.gaugeArcs = new TachoBase.GaugeArcs(
              new TachoBase.Arc(
                  arcRadius,
                  // Will be overridden later, so default to no arc
                  new TachoBase.Range(this.angleRange.lower, this.angleRange.lower),
                  new TachoBase.Style(this.colourScheme.ACCENT.secondary, arcWidth)
              ),
              new TachoBase.Arc(
                  arcRadius,
                  new TachoBase.Range(redlineAngle, this.angleRange.upper),
                  new TachoBase.Style("red", arcWidth)
              )
          );
          return this;
        }
      }

      this.gauge = new TachoVelocityGaugeBuilder(this.radius, this.colourScheme, this.RPM_ANGLE_RANGE)
          .createPointer(
              VelocityTachoApp.RADII.RPM_DIAL_RADIUS,
              this.CENTRE,
              this.radius - VelocityTachoApp.WIDTH.RPM_ARC
          )
          // +1 to include 0RPM
          .createTicks(
              Math.floor(this.redLineRPM / VelocityTachoApp.TICK_SPACING) + 1,
              VelocityTachoApp.TICK_SPACING,
              VelocityTachoApp.WIDTH.RPM_ARC
          )
          .createArcs(
              VelocityTachoApp.RADII.RPM_ARC,
              VelocityTachoApp.WIDTH.RPM_ARC,
              this.getRedlineRPMAngle()
          )
          // +1 to include 0
          .createNumbers(this.redLineRPM/1000 + 1, 20, VelocityTachoApp.RADII.RPM_NUMBERS, 1)
          .setCentre(this.CENTRE)
          .build();
    }

    /**
     * @Override
     * @inheritDoc
     * This includes both backgrounds for the speedometer as well as the divider between the
     * RPM gauge and inner part. This also draws the background parts of the RPM (see
     * setupRPM()) and draws the gear in the centre (see drawGear())
     */
    drawBackgroundCanvas() {
      super.drawBackgroundCanvas();
      let ctx = this.BG_CANVAS.get2DContext();

      // The background wedge for the last quarter not filled by the RPM
      new TachoBase.Wedge(
          VelocityTachoApp.RADII.BASE, this.RPM_ANGLE_RANGE.invert(),
          TachoBase.Style.noStrokeFill(this.colourScheme.BASE), this.CENTRE
      ).draw(ctx);
      // The inner circle part
      new TachoBase.Wedge(
          VelocityTachoApp.RADII.MIDDLE,
          this.RPM_ANGLE_RANGE,
          TachoBase.Style.noStrokeFill(this.colourScheme.BACKGROUND), this.CENTRE
      ).draw(ctx);

      this.setupRPM();
      this.drawGear();
      this.drawSpeedBackground();
      this.drawParkingBrake();
    }

    /**
     * Setups and draws the RPM gauge's background canvas objects.
     * This draws the dividing arcs as well as the backgrounds (such as redline RPM) for the RPM.
     * This also draws the inner parts of the RPM pointer as well as generating the RPM ticks
     */
    setupRPM() {
      const STROKE_WIDTH = 10;
      const RADIUS = VelocityTachoApp.RADII.BG_INNER - STROKE_WIDTH/2;

      let ctx = this.BG_CANVAS.get2DContext();
      let redlineRPMAngle = this.getRedlineRPMAngle();

      // Draw interior arc between RPM background and the interior
      new TachoBase.Arc(
          RADIUS,
          this.RPM_ANGLE_RANGE,
          new TachoBase.Style(this.colourScheme.ACCENT.primary, STROKE_WIDTH),
          this.CENTRE
      ).draw(ctx);
      // The background arc to the RPM numbers
      new TachoBase.Arc(
          VelocityTachoApp.RADII.RPM_ARC_BG,
          this.RPM_ANGLE_RANGE,
          new TachoBase.Style(this.colourScheme.BASE, VelocityTachoApp.WIDTH.RPM_BG + 1),
          this.CENTRE
      ).draw(ctx);

      // Draw non-redline background RPM
      new TachoBase.Arc(
          VelocityTachoApp.RADII.RPM_ARC,
          new TachoBase.Range(this.RPM_ANGLE_RANGE.lower, redlineRPMAngle),
          new TachoBase.Style(this.colourScheme.ACCENT.primary, VelocityTachoApp.WIDTH.RPM_ARC),
          this.CENTRE
      ).draw(ctx);

      // Draw the pointer dial as this only needs to be drawn once
      this.gauge.pointer.pointerDial.draw(ctx);
    }

    /**
     * Draws the gear number in the middle of the tachometer
     */
    drawGear() {
      let gearName = TachoBase.TachometerApp.gearIndexToName(this.currGearIndex);
      // Base font size is 40; if there is more >1 digit on the gear then decrease to fit
      let fontSize = 40 / (1 + (gearName.length - 1) * 0.5);

      let ctx = this.CANVAS.get2DContext();

      this.drawPrimaryText(ctx, gearName, fontSize, this.CENTRE);
      TachoBase.BaseApp.resetFilter(ctx);
    }

    /**
     * Draws the background for the oil and fuel bars
     */
    drawBarBackground() {
      let radii = VelocityTachoApp.RADII;
      const DIFF = (radii.TEMP - radii.FUEL)/2;
      // The padding either side of the bars
      const PADDING = VelocityTachoApp.WIDTH.INNER_STROKE;
      const MID_RADIUS = radii.TEMP - DIFF - PADDING;

      const WIDTH = DIFF + VelocityTachoApp.WIDTH.BARS*2 + PADDING;

      let range = VelocityTachoApp.getBarRange();
      // Pad out each end
      range.pad(0.2);

      new TachoBase.Arc(MID_RADIUS, range, new TachoBase.Style("black", WIDTH, false), this.CENTRE)
          .draw(this.CANVAS.get2DContext());
    }

    /**
     * Draws the background polygon for the speedometer
     */
    drawSpeedBackground() {
      let ctx = this.BG_CANVAS.get2DContext();
      outerThis.drawRoundRectRight(ctx, this, 170);

      // Apply lighting glow
      if (this.lightsState > 0) {
        // Green for low beam, blue for high beam
        let colour = this.lightsState === 1 ? "green" : "rgb(0, 100, 255)";
        ctx.filter = "drop-shadow(-8px -8px 8px " + colour + ")";
      }

      ctx.fill();
      TachoBase.BaseApp.resetFilter(ctx);
    }

    /**
     * @returns {Range} the angle range for the parking brake's background polygonal shape
     */
    getPBrakePolygonAngles() {
      return new TachoBase.Range(
          this.RPM_ANGLE_RANGE.lower * 0.475,
          this.RPM_ANGLE_RANGE.lower * 0.825
      );
    }

    /**
     * Draws the background polygon of the parking brake text
     */
    drawParkingBrake() {
      let ctx = this.BG_CANVAS.get2DContext();
      ctx.beginPath();

      new TachoBase.RoundCornerRectangleBR(this.radius, this.getPBrakePolygonAngles(), this.CENTRE, 170)
          .drawContinue(ctx);

      ctx.lineWidth = 1;
      ctx.fillStyle = "black";
      ctx.fill();

      this.drawParkingBrakeText();
    }

    /**
     * Draws the parking brake text, coloured to indicate whether it is applied or not
     * Draws the parking brake text, coloured to indicate whether it is applied or not
     */
    drawParkingBrakeText() {
      let colour = this.pBrakeEngaged ? "orange" : "rgb(24, 24, 24)";

      let centrePos = TachoBase.getRelCircularPoint(this.CENTRE, VelocityTachoApp.RADII.RPM_ARC, this.getPBrakePolygonAngles().getLinearRatio(0.5));
      // Manual tuning to make it look more central
      centrePos.addY(8);
      centrePos.addX(-20);

      let ctx = this.BG_CANVAS.get2DContext();

      // Apply glow
      if (this.pBrakeEngaged)
        ctx.filter = "drop-shadow(0 0 8px " + colour + ")";

      new TachoBase.Text(new TachoBase.Font(this.FONT_NAME, 15, true), "PARK", centrePos, colour)
          .draw(ctx);
      TachoBase.BaseApp.resetFilter(ctx);
    }

    /**
     * @Override
     * @inheritDoc
     * Draws the RPM gauge and the interior
     */
    draw() {
      super.draw();
      this.drawBars()
      this.drawRPM();
      this.drawInterior();
    }

    /**
     * Draws the interior parts of the tachometer. This includes the speedometer
     */
    drawInterior() {
      // Re-draw background canvas if a redraw is required
      if (this.needsRedraw) {
        this.drawBackgroundCanvas();
        this.needsRedraw = false;
      }

      this.drawGear();

      let convertedSpeed = this.UiUnits.speed(this.speed, this.unit);
      this.drawSpeed(Math.floor(convertedSpeed.val), convertedSpeed.unit);
    }

    /**
     * Draws the temperature and fuel bars in the centre of the UI as well as the bar background
     */
    drawBars() {
      this.drawBarBackground();
      this.drawOilTempBar();
      this.drawFuelBar();
    }

    drawFuelBar() {
      // Constrain to [0,1]
      let perc = Math.min(1, Math.max(0, this.fuel / 100));
      this.fuelBar.setPercentageAlong(perc);
      this.fuelBar.draw(this.CANVAS.get2DContext());
    }

    /**
     * Draws the oil temperature arc bar
     */
    drawOilTempBar() {
      // The maximum temperature of oil shown (as used in BeamNG's UI)
      const MAX_TEMP = 120;
      // Constrain to [0,1]
      let perc = Math.min(1, Math.max(0, this.engineTemp / MAX_TEMP));

      this.oilTempBar.setPercentageAlong(perc);
      this.oilTempBar.draw(this.CANVAS.get2DContext());
    }

    /**
     * Creates the oil temperature bar
     */
    createOilTempBar() {
      return this.createArcBar(VelocityTachoApp.RADII.TEMP, "#ff3333");
    }

    /**
     * Creates the fuel bar
     */
    createFuelBar() {
      return this.createArcBar(VelocityTachoApp.RADII.FUEL, "#0073ff");
    }

    /**
     * Creates an arc bar in this theme's style
     * @param {number} radius - the arc's radius
     * @param {string} colour - the arc's colour
     * @returns {ArcBar} the ArcBar created
     */
    createArcBar(radius, colour) {
      return new TachoBase.ArcBar(
          radius,
          VelocityTachoApp.getBarRange(),
          new TachoBase.Style(colour, VelocityTachoApp.WIDTH.BARS, false),
          this.CENTRE
      );
    }

    /**
     * @returns {Range} the bar range for the oil and fuel bars
     */
    static getBarRange() {
      return new TachoBase.Range(0.65*Math.PI, 1.35*Math.PI);
    }

    /**
     * Draws the speedometer and its units
     * @param {number} speed - the speed of the vehicle
     * @param {string} unit - the unit {@param speed} is in
     */
    drawSpeed(speed, unit) {
      // The spacing between pieces of text and the space between the top of the speed-box and
      // the first piece of text
      const SPACING = 5;
      const SPEED_FONT_SIZE = 40;
      const UNIT_FONT_SIZE = 20;
      let centrePos = TachoBase.getRelCircularPoint(this.CENTRE, VelocityTachoApp.RADII.BG_INNER, outerThis.roundRectRightAngle(this).getLinearRatio(0.5));

      let ctx = this.CANVAS.get2DContext();

      // Centre the text
      centrePos.addY(SPACING * 3);
      this.drawPrimaryText(ctx, speed.toString(), SPEED_FONT_SIZE, centrePos);
      // Account for the size of the speed text to ensure no-overlap
      centrePos.addY(SPEED_FONT_SIZE / 2 + SPACING);
      this.drawPrimaryText(ctx, unit, UNIT_FONT_SIZE, centrePos);
    }

    /**
     * Draws the RPM gauge. This includes the RPM ticks, pointer and current RPM arc
     */
    drawRPM() {
      let currRPMAngle = this.RPM_ANGLE_RANGE.getLinearRatio(this.currRPM / this.redLineRPM);
      this.gauge.setAngle(currRPMAngle);
      this.gauge.draw(this.CANVAS.get2DContext(), this.currRPM);
    }

    /**
     * Draws centred text in the colour scheme's primary text colour at the position specified
     * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
     * @param {string} text - the text to write
     * @param {number} fontSize - the font size in px
     * @param {Vector} position - the position to set the centre of the text to
     */
    drawPrimaryText(ctx, text, fontSize, position) {
      new TachoBase.Text(
          new TachoBase.Font(this.FONT_NAME, fontSize, true),
          text,
          position,
          this.colourScheme.TEXT.primary
      ).draw(ctx);
    }
  }


  const VFI_WIDTH = {
    // The current boost arc that goes around the outside
    BOOST_ARC: 20,
    // The width of each tick
    TICK_WIDTH: 2
  }

  const VFI_RADII = {
    BASE: 75,
    DIAL_RADIUS: 10
  };
  VFI_RADII.BOOST_ARC = VFI_RADII.BASE - VFI_WIDTH.BOOST_ARC/2;
  VFI_RADII.MIDDLE = 40;
  VFI_RADII.BOOST_NUMBERS = (VFI_RADII.BASE - VFI_WIDTH.BOOST_ARC + VFI_RADII.MIDDLE)/2;

  VFI_WIDTH.BOOST_NUMBERS = VFI_RADII.BASE - VFI_WIDTH.BOOST_ARC - VFI_RADII.MIDDLE;

  /**
   * The concrete class representation of this app
   */
  class VelocityBoostApp extends TachoBase.BoostApp {
    // The width of stroke effects for various components
    static WIDTH = VFI_WIDTH;

    // The centre radius values for various components
    // Any stroke effect adds thickness to a radius
    static RADII = VFI_RADII;

    // The number of ticks between two numbers on the gauge
    static NUM_TICKS_PER_GAP = 3;
    // The number of numbers on the gauge to display
    static NUM_NUMBERS = 5;

    constructor(id, document, bngApi, UiUnits) {
      super(id, outerThis.GAUGE_RANGE, outerThis.FONT_NAME, VelocityBoostApp.RADII.BASE, document, bngApi, UiUnits);
    }

    /**
     * @Override
     * @inheritDoc
     */
    createGauge() {
      // Create concrete class of base class builder
      class TachoVelocityGaugeBuilder extends outerThis.GaugeBuilder {
        constructor(baseRadius, colourScheme, angleRange) {
          super(baseRadius, colourScheme, angleRange);
        }

        /**
         * Creates the GaugeArcs instance for the gauge
         * @param arcRadius
         * @param arcWidth
         * @param redlineAngle
         * @returns {TachoVelocityGaugeBuilder} the GaugeBuilder instance
         */
        createArcs(arcRadius, arcWidth, redlineAngle) {
          this.gaugeArcs = new TachoBase.GaugeArcs(
              new TachoBase.Arc(
                  arcRadius,
                  // Will be overridden later, so default to no arc
                  new TachoBase.Range(this.angleRange.lower, this.angleRange.lower),
                  new TachoBase.Style(this.colourScheme.ACCENT.secondary, arcWidth)
              )
          );
          return this;
        }
      }


      let unitMaxBoost = this.UiUnits.pressure(this.maxBoost, this.unit).val;
      let floorMaxBoost = Math.floor(unitMaxBoost + 1);
      // Increment between numbers on the gauge
      let numberInc = floorMaxBoost / Math.floor(VelocityBoostApp.NUM_NUMBERS / 2);

      let numTicks = VelocityBoostApp.NUM_TICKS_PER_GAP * (VelocityBoostApp.NUM_NUMBERS-1)
                     + VelocityBoostApp.NUM_NUMBERS;
      let tickSpacing = numberInc / (VelocityBoostApp.NUM_TICKS_PER_GAP + 1);

      this.gauge = new TachoVelocityGaugeBuilder(
          VelocityBoostApp.RADII.BASE,
          this.colourScheme,
          this.BOOST_ANGLE_RANGE
      )
          .createPointer(
              VelocityBoostApp.RADII.DIAL_RADIUS,
              this.CENTRE,
              this.radius - VelocityBoostApp.WIDTH.BOOST_ARC
          )
          // +1 to include 0 boost
          .createTicks(numTicks, tickSpacing, VelocityBoostApp.WIDTH.BOOST_ARC, -unitMaxBoost)
          .createArcs(VelocityBoostApp.RADII.BOOST_ARC, VelocityBoostApp.WIDTH.BOOST_ARC)
          .createNumbers(
              VelocityBoostApp.NUM_NUMBERS,
              10,
              VelocityBoostApp.RADII.BOOST_NUMBERS,
              numberInc,
              -floorMaxBoost
          )
          .setCentre(this.CENTRE)
          .build();
    }

    /**
     * Draws/updates the background canvas
     */
    drawBackgroundCanvas() {
      super.drawBackgroundCanvas();
      const STROKE_WIDTH = 6;
      const RADIUS = VelocityBoostApp.RADII.MIDDLE - STROKE_WIDTH/2;

      let ctx = this.BG_CANVAS.get2DContext();

      // The inner circle part
      new TachoBase.Wedge(VelocityBoostApp.RADII.MIDDLE, this.BOOST_ANGLE_RANGE, TachoBase.Style.noStrokeFill(this.colourScheme.BACKGROUND), this.CENTRE)
          .draw(ctx);
      // Draw interior arc between boost arc background and the interior
      new TachoBase.Arc(RADIUS, this.BOOST_ANGLE_RANGE, new TachoBase.Style(this.colourScheme.ACCENT.primary, STROKE_WIDTH), this.CENTRE)
          .draw(ctx);

      // The wedge not covered by the gauge area
      new TachoBase.Wedge(VelocityBoostApp.RADII.BASE, this.BOOST_ANGLE_RANGE.invert(), TachoBase.Style.noStrokeFill(this.colourScheme.BASE), this.CENTRE)
          .draw(ctx);

      // The background arc to the RPM numbers
      new TachoBase.Arc(VelocityBoostApp.RADII.BOOST_NUMBERS, this.BOOST_ANGLE_RANGE, new TachoBase.Style(this.colourScheme.BASE, VelocityBoostApp.WIDTH.BOOST_NUMBERS+1), this.CENTRE)
          .draw(ctx);
      // Draw background arc to boost gauge
      // Note: -1 to reduce overlap when the other canvas draws over it
      new TachoBase.Arc(VelocityBoostApp.RADII.BOOST_ARC, this.BOOST_ANGLE_RANGE, new TachoBase.Style(this.colourScheme.ACCENT.primary, VelocityBoostApp.WIDTH.BOOST_ARC), this.CENTRE)
          .draw(ctx);

      // Draw the pointer dial as this only needs to be drawn once
      this.gauge.pointer.pointerDial.draw(ctx);

      this.drawUnit();
    }

    /**
     * Draws the boost's unit
     */
    drawUnit() {
      this.drawUnitBackground();
      this.drawUnitText();
    }

    /**
     * Draws the background of the unit area
     */
    drawUnitBackground() {
      let ctx = this.BG_CANVAS.get2DContext();
      outerThis.drawRoundRectRight(ctx, this, 90);
      ctx.fill();
    }

    /**
     * Draws the unit's name in text form
     */
    drawUnitText() {
      const FONT_SIZE = 20;
      const SPACING = 10;

      let ctx = this.BG_CANVAS.get2DContext();

      let centrePos = TachoBase.getRelCircularPoint(this.CENTRE, VelocityBoostApp.RADII.BOOST_NUMBERS, this.BOOST_ANGLE_RANGE.lower * 0.275);
      centrePos.addY(SPACING);

      new TachoBase.Text(new TachoBase.Font(this.FONT_NAME, FONT_SIZE, true), this.unit, centrePos, this.colourScheme.TEXT.primary)
          .draw(ctx);
    }

    /**
     * @Override
     * @inheritDoc
     */
    draw() {
      super.draw();
      let ctx = this.CANVAS.get2DContext();

      let boost = this.UiUnits.pressure(this.currBoost, this.unit).val;

      // Normalise from range [-1,1] to [0,1]
      let normBoost = (this.currBoost/this.maxBoost + 1)/2;
      let currBoostAngle = this.BOOST_ANGLE_RANGE.getLinearRatio(normBoost);

      this.gauge.setAngle(currBoostAngle);
      this.gauge.draw(ctx, boost);
    }
  }


  /**
   * Returns the the angle range for the middle-right rounded rectangle's arc
   * @param {BaseApp} app - the app intending to draw the round rectangle polygon using this angle
   * @returns {Range}
   * @see drawRoundRectRight()
   */
  this.roundRectRightAngle = function(app) {
    let angleRange = app.gauge.gaugeTicks.angleRange;
    return new TachoBase.Range(angleRange.lower * 0.125, angleRange.lower * 0.425);
  }

  /**
   * Draws (but does not complete) a rounded rectangle at the mid-right position of the UI
   * @param {CanvasRenderingContext2D} ctx - the rendering context of the canvas to draw on
   * @param {BaseApp} app - the app drawing this polygon
   * @param {number} leftX - the left-most x coordinate
   */
  this.drawRoundRectRight = function(ctx, app, leftX) {
    ctx.beginPath();
    new TachoBase.RoundCornerRectangleBR(
        app.radius, this.roundRectRightAngle(app), app.CENTRE, leftX, null
    ).drawContinue(ctx);
    ctx.lineWidth = 1;
    ctx.fillStyle = "black";
  }


  this.TachoBase = TachoBase;
  this.GaugeBuilder = GaugeBuilder;
  this.VelocityTachoApp = VelocityTachoApp;
  this.VelocityBoostApp = VelocityBoostApp;

  // The name of the font to use
  this.FONT_NAME = "Arial Black";
  // The angle range of the gauges
  this.GAUGE_RANGE = new TachoBase.Range(0.5*Math.PI, 2*Math.PI);


  return this;
}]);